@@ -56,6 +56,9 @@ struct _FlashbackMonitorManagerPrivate
   int                    rr_event_base;
   int                    rr_error_base;
 
+  gboolean               has_randr13;
+  gboolean               has_randr15;
+
   MetaDBusDisplayConfig *display_config;
 };
 
@@ -70,6 +73,96 @@ enum
 
 static GParamSpec *object_properties[N_PROPERTIES] = { NULL, };
 
+static void
+add_monitor (FlashbackMonitorManager *manager,
+             MetaMonitorInfo         *monitor)
+{
+#ifdef HAVE_XRANDR15
+  XRRMonitorInfo *m;
+  int o;
+  Atom name;
+  char name_buf[40];
+
+  if (manager->priv->has_randr15 == FALSE)
+    return;
+
+  if (monitor->n_outputs <= 1)
+    return;
+
+  if (monitor->outputs[0]->product)
+    snprintf (name_buf, 40, "%s-%d", monitor->outputs[0]->product, monitor->outputs[0]->tile_info.group_id);
+  else
+    snprintf (name_buf, 40, "Tiled-%d", monitor->outputs[0]->tile_info.group_id);
+
+  name = XInternAtom (manager->priv->xdisplay, name_buf, False);
+  monitor->monitor_winsys_xid = name;
+  m = XRRAllocateMonitor (manager->priv->xdisplay, monitor->n_outputs);
+  if (!m)
+    return;
+
+  m->name = name;
+  m->primary = monitor->is_primary;
+  m->automatic = True;
+
+  for (o = 0; o < monitor->n_outputs; o++)
+    {
+      MetaOutput *output = monitor->outputs[o];
+      m->outputs[o] = output->winsys_id;
+    }
+
+  XRRSetMonitor (manager->priv->xdisplay,
+                 DefaultRootWindow (manager->priv->xdisplay),
+                 m);
+
+  XRRFreeMonitors (m);
+#endif
+}
+
+static void
+remove_monitor (FlashbackMonitorManager *manager,
+                int                      monitor_winsys_xid)
+{
+#ifdef HAVE_XRANDR15
+  if (manager->priv->has_randr15 == FALSE)
+    return;
+
+  XRRDeleteMonitor (manager->priv->xdisplay,
+                    DefaultRootWindow (manager->priv->xdisplay),
+                    monitor_winsys_xid);
+#endif
+}
+
+static void
+init_monitors (FlashbackMonitorManager *manager)
+{
+#ifdef HAVE_XRANDR15
+  XRRMonitorInfo *m;
+  int n, i;
+
+  if (manager->priv->has_randr15 == FALSE)
+    return;
+
+  /* delete any tiled monitors setup, as mutter will want to recreate
+     things in its image */
+  m = XRRGetMonitors (manager->priv->xdisplay,
+                      DefaultRootWindow (manager->priv->xdisplay),
+                      FALSE, &n);
+
+  if (n == -1)
+    return;
+
+  for (i = 0; i < n; i++)
+    {
+      if (m[i].noutput > 1)
+        XRRDeleteMonitor (manager->priv->xdisplay,
+                          DefaultRootWindow (manager->priv->xdisplay),
+                          m[i].name);
+    }
+
+  XRRFreeMonitors (m);
+#endif
+}
+
 static GdkFilterReturn
 filter_func (GdkXEvent *xevent,
              GdkEvent  *event,
@@ -134,9 +227,10 @@ meta_monitor_transform_to_xrandr (MetaMonitorTransform transform)
       return RR_Reflect_X | RR_Rotate_180;
     case META_MONITOR_TRANSFORM_FLIPPED_270:
       return RR_Reflect_X | RR_Rotate_270;
+    default:
+      g_assert_not_reached ();
+      break;
     }
-
-  g_assert_not_reached ();
 }
 
 static MetaMonitorTransform
@@ -313,11 +407,11 @@ output_set_presentation_xrandr (FlashbackMonitorManagerPrivate *priv,
 
   value = presentation;
   atom = XInternAtom (priv->xdisplay, "_GNOME_FLASHBACK_PRESENTATION_OUTPUT", False);
-  XRRChangeOutputProperty (priv->xdisplay,
-                           (XID) output->winsys_id,
-                           atom,
-                           XA_CARDINAL, 32, PropModeReplace,
-                           (unsigned char*) &value, 1);
+  xcb_randr_change_output_property (XGetXCBConnection (priv->xdisplay),
+                                    (XID) output->winsys_id,
+                                    atom, XCB_ATOM_CARDINAL, 32,
+                                    XCB_PROP_MODE_REPLACE,
+                                    1, &value);
 }
 
 static gboolean
@@ -352,6 +446,54 @@ out:
   return value;
 }
 
+static gboolean
+output_get_supports_underscanning_xrandr (FlashbackMonitorManagerPrivate *priv,
+                                          MetaOutput                     *output)
+{
+  Atom atom, actual_type;
+  int actual_format, i;
+  unsigned long nitems, bytes_after;
+  unsigned char *buffer;
+  XRRPropertyInfo *property_info;
+  Atom *values;
+  gboolean supports_underscanning = FALSE;
+
+  atom = XInternAtom (priv->xdisplay, "underscan", False);
+  XRRGetOutputProperty (priv->xdisplay,
+                        (XID)output->winsys_id,
+                        atom,
+                        0, G_MAXLONG, False, False, XA_ATOM,
+                        &actual_type, &actual_format,
+                        &nitems, &bytes_after, &buffer);
+
+  if (actual_type != XA_ATOM || actual_format != 32 || nitems < 1)
+    {
+      XFree (buffer);
+      return FALSE;
+    }
+
+  property_info = XRRQueryOutputProperty (priv->xdisplay,
+                                          (XID) output->winsys_id,
+                                          atom);
+  values = (Atom *) property_info->values;
+
+  for (i = 0; i < property_info->num_values; i++)
+    {
+      /* The output supports underscanning if "on" is a valid value
+       * for the underscan property.
+       */
+      char *name = XGetAtomName (priv->xdisplay, values[i]);
+      if (strcmp (name, "on") == 0)
+        supports_underscanning = TRUE;
+
+      XFree (name);
+    }
+
+  XFree (property_info);
+
+  return supports_underscanning;
+}
+
 static int
 normalize_backlight (MetaOutput *output,
                      int         hw_value)
@@ -399,6 +541,7 @@ output_get_backlight_limits_xrandr (FlashbackMonitorManagerPrivate *priv,
   Atom atom;
   xcb_connection_t *xcb_conn;
   xcb_randr_query_output_property_reply_t *reply;
+  int32_t *values;
 
   atom = XInternAtom (priv->xdisplay, "Backlight", False);
 
@@ -419,7 +562,7 @@ output_get_backlight_limits_xrandr (FlashbackMonitorManagerPrivate *priv,
       goto out;
     }
 
-  int32_t *values = xcb_randr_query_output_property_valid_values (reply);
+  values = xcb_randr_query_output_property_valid_values (reply);
   output->backlight_min = values[0];
   output->backlight_max = values[1];
 
@@ -427,6 +570,42 @@ out:
   free (reply);
 }
 
+static void
+output_get_tile_info (FlashbackMonitorManagerPrivate *priv,
+                      MetaOutput                     *output)
+{
+  Atom tile_atom;
+  unsigned char *prop;
+  unsigned long nitems, bytes_after;
+  int actual_format;
+  Atom actual_type;
+
+  if (priv->has_randr15 == FALSE)
+    return;
+
+  tile_atom = XInternAtom (priv->xdisplay, "TILE", FALSE);
+  XRRGetOutputProperty (priv->xdisplay, output->winsys_id,
+                        tile_atom, 0, 100, False,
+                        False, AnyPropertyType,
+                        &actual_type, &actual_format,
+                        &nitems, &bytes_after, &prop);
+
+  if (actual_type == XA_INTEGER && actual_format == 32 && nitems == 8)
+    {
+      long *values = (long *)prop;
+      output->tile_info.group_id = values[0];
+      output->tile_info.flags = values[1];
+      output->tile_info.max_h_tiles = values[2];
+      output->tile_info.max_v_tiles = values[3];
+      output->tile_info.loc_h_tile = values[4];
+      output->tile_info.loc_v_tile = values[5];
+      output->tile_info.tile_w = values[6];
+      output->tile_info.tile_h = values[7];
+    }
+
+  XFree (prop);
+}
+
 static gboolean
 output_get_hotplug_mode_update (FlashbackMonitorManagerPrivate *priv,
                                 MetaOutput                     *output)
@@ -617,11 +796,105 @@ get_xmode_name (XRRModeInfo *xmode)
 }
 
 static gboolean
-gdk_rectangle_equal (const GdkRectangle *src1,
-                     const GdkRectangle *src2)
+gdk_rectangle_contains_rect (const GdkRectangle *outer_rect,
+                             const GdkRectangle *inner_rect)
 {
-  return ((src1->x == src2->x) && (src1->y == src2->y) &&
-          (src1->width == src2->width) && (src1->height == src2->height));
+  return inner_rect->x                      >= outer_rect->x &&
+         inner_rect->y                      >= outer_rect->y &&
+         inner_rect->x + inner_rect->width  <= outer_rect->x + outer_rect->width &&
+         inner_rect->y + inner_rect->height <= outer_rect->y + outer_rect->height;
+}
+
+/*
+ * rules for constructing a tiled monitor
+ * 1. find a tile_group_id
+ * 2. iterate over all outputs for that tile group id
+ * 3. see if output has a crtc and if it is configured for the tile size
+ * 4. calculate the total tile size
+ * 5. set tile finished size
+ * 6. check for more tile_group_id
+*/
+static void
+construct_tile_monitor (FlashbackMonitorManager *manager,
+                        GArray                  *monitor_infos,
+                        guint32                  tile_group_id)
+{
+  MetaMonitorInfo info;
+  unsigned i;
+
+  for (i = 0; i < monitor_infos->len; i++)
+    {
+      MetaMonitorInfo *pinfo = &g_array_index (monitor_infos, MetaMonitorInfo, i);
+
+      if (pinfo->tile_group_id == tile_group_id)
+        return;
+    }
+
+  /* didn't find it */
+  info.number = monitor_infos->len;
+  info.tile_group_id = tile_group_id;
+  info.is_presentation = FALSE;
+  info.refresh_rate = 0.0;
+  info.width_mm = 0;
+  info.height_mm = 0;
+  info.is_primary = FALSE;
+  info.rect.x = INT_MAX;
+  info.rect.y = INT_MAX;
+  info.rect.width = 0;
+  info.rect.height = 0;
+  info.winsys_id = 0;
+  info.n_outputs = 0;
+  info.monitor_winsys_xid = 0;
+
+  for (i = 0; i < manager->n_outputs; i++)
+    {
+      MetaOutput *output = &manager->outputs[i];
+
+      if (!output->tile_info.group_id)
+        continue;
+
+      if (output->tile_info.group_id != tile_group_id)
+        continue;
+
+      if (!output->crtc)
+        continue;
+
+      if (output->crtc->rect.width != (int)output->tile_info.tile_w ||
+          output->crtc->rect.height != (int)output->tile_info.tile_h)
+        continue;
+
+      if (output->tile_info.loc_h_tile == 0 && output->tile_info.loc_v_tile == 0)
+        {
+          info.refresh_rate = output->crtc->current_mode->refresh_rate;
+          info.width_mm = output->width_mm;
+          info.height_mm = output->height_mm;
+          info.winsys_id = output->winsys_id;
+        }
+
+      /* hack */
+      if (output->crtc->rect.x < info.rect.x)
+        info.rect.x = output->crtc->rect.x;
+      if (output->crtc->rect.y < info.rect.y)
+        info.rect.y = output->crtc->rect.y;
+
+      if (output->tile_info.loc_h_tile == 0)
+        info.rect.height += output->tile_info.tile_h;
+
+      if (output->tile_info.loc_v_tile == 0)
+        info.rect.width += output->tile_info.tile_w;
+
+      if (info.n_outputs > META_MAX_OUTPUTS_PER_MONITOR)
+        continue;
+
+      info.outputs[info.n_outputs++] = output;
+    }
+
+  /* if we don't have a winsys id, i.e. we haven't found tile 0,0
+     don't try and add this to the monitor infos */
+  if (!info.winsys_id)
+    return;
+
+  g_array_append_val (monitor_infos, info);
 }
 
 /*
@@ -644,6 +917,16 @@ make_logical_config (FlashbackMonitorManager *manager)
      for each of them, unless they reference a rectangle that
      is already there.
   */
+
+  /* for tiling we need to work out how many tiled outputs there are */
+  for (i = 0; i < manager->n_outputs; i++)
+    {
+      MetaOutput *output = &manager->outputs[i];
+
+      if (output->tile_info.group_id)
+        construct_tile_monitor (manager, monitor_infos, output->tile_info.group_id);
+    }
+
   for (i = 0; i < manager->n_crtcs; i++)
     {
       MetaCRTC *crtc = &manager->crtcs[i];
@@ -655,7 +938,7 @@ make_logical_config (FlashbackMonitorManager *manager)
       for (j = 0; j < monitor_infos->len; j++)
         {
           MetaMonitorInfo *info = &g_array_index (monitor_infos, MetaMonitorInfo, j);
-          if (gdk_rectangle_equal (&crtc->rect, &info->rect))
+          if (gdk_rectangle_contains_rect (&crtc->rect, &info->rect))
             {
               crtc->logical_monitor = info;
               break;
@@ -667,7 +950,9 @@ make_logical_config (FlashbackMonitorManager *manager)
           MetaMonitorInfo info;
 
           info.number = monitor_infos->len;
+          info.tile_group_id = 0;
           info.rect = crtc->rect;
+          info.refresh_rate = crtc->current_mode->refresh_rate;
           info.is_primary = FALSE;
           /* This starts true because we want
              is_presentation only if all outputs are
@@ -678,6 +963,9 @@ make_logical_config (FlashbackMonitorManager *manager)
           info.in_fullscreen = -1;
           info.winsys_id = 0;
 
+          info.n_outputs = 0;
+          info.monitor_winsys_xid = 0;
+
           g_array_append_val (monitor_infos, info);
 
           crtc->logical_monitor = &g_array_index (monitor_infos, MetaMonitorInfo,
@@ -699,6 +987,9 @@ make_logical_config (FlashbackMonitorManager *manager)
       if (output->crtc == NULL)
         continue;
 
+      if (output->tile_info.group_id)
+        continue;
+
       /* We must have a logical monitor on every CRTC at this point */
       g_assert (output->crtc->logical_monitor != NULL);
 
@@ -707,6 +998,12 @@ make_logical_config (FlashbackMonitorManager *manager)
       info->is_primary = info->is_primary || output->is_primary;
       info->is_presentation = info->is_presentation && output->is_presentation;
 
+      info->width_mm = output->width_mm;
+      info->height_mm = output->height_mm;
+
+      info->outputs[0] = output;
+      info->n_outputs = 1;
+
       if (output->is_primary || info->winsys_id == 0)
         info->winsys_id = output->winsys_id;
 
@@ -716,6 +1013,9 @@ make_logical_config (FlashbackMonitorManager *manager)
 
   manager->n_monitor_infos = monitor_infos->len;
   manager->monitor_infos = (void*)g_array_free (monitor_infos, FALSE);
+
+  for (i = 0; i < manager->n_monitor_infos; i++)
+    add_monitor (manager, &manager->monitor_infos[i]);
 }
 
 static void
@@ -891,7 +1191,11 @@ read_current_config (FlashbackMonitorManager *manager)
   manager->screen_width = WidthOfScreen (screen);
   manager->screen_height = HeightOfScreen (screen);
 
-  resources = XRRGetScreenResourcesCurrent (priv->xdisplay, DefaultRootWindow (priv->xdisplay));
+  if (priv->has_randr13)
+    resources = XRRGetScreenResourcesCurrent (priv->xdisplay, DefaultRootWindow (priv->xdisplay));
+  else
+    resources = XRRGetScreenResources (priv->xdisplay, DefaultRootWindow (priv->xdisplay));
+
   if (!resources)
     return;
 
@@ -959,7 +1263,7 @@ read_current_config (FlashbackMonitorManager *manager)
 
       meta_output = &manager->outputs[n_actual_outputs];
 
-      if (output->connection != RR_Disconnected)
+      if (output->connection != RR_Disconnected && output->nmode != 0)
         {
           GBytes *edid;
 
@@ -977,6 +1281,8 @@ read_current_config (FlashbackMonitorManager *manager)
           meta_output->suggested_y = output_get_suggested_y (priv, meta_output);
           meta_output->connector_type = output_get_connector_type (priv, meta_output);
 
+          output_get_tile_info (priv, meta_output);
+
           meta_output->n_modes = output->nmode;
           meta_output->modes = g_new0 (MetaMonitorMode *, meta_output->n_modes);
 
@@ -1034,6 +1340,7 @@ read_current_config (FlashbackMonitorManager *manager)
           meta_output->is_primary = ((XID)meta_output->winsys_id == primary_output);
           meta_output->is_presentation = output_get_presentation_xrandr (priv, meta_output);
           meta_output->is_underscanning = output_get_underscanning_xrandr (priv, meta_output);
+          meta_output->supports_underscanning = output_get_supports_underscanning_xrandr (priv, meta_output);
           output_get_backlight_limits_xrandr (priv, meta_output);
 
         if (!(meta_output->backlight_min == 0 && meta_output->backlight_max == 0))
@@ -1199,6 +1506,8 @@ static void
 flashback_monitor_manager_init (FlashbackMonitorManager *manager)
 {
   FlashbackMonitorManagerPrivate *priv;
+  int major_version;
+  int minor_version;
 
   priv = flashback_monitor_manager_get_instance_private (manager);
   manager->priv = priv;
@@ -1215,6 +1524,21 @@ flashback_monitor_manager_init (FlashbackMonitorManager *manager)
   XRRSelectInput (priv->xdisplay, DefaultRootWindow (priv->xdisplay),
                   RRScreenChangeNotifyMask | RRCrtcChangeNotifyMask |
                   RROutputPropertyNotifyMask);
+
+  priv->has_randr13 = FALSE;
+  priv->has_randr15 = FALSE;
+
+  XRRQueryVersion (priv->xdisplay, &major_version, &minor_version);
+
+  if (major_version > 1 || (major_version == 1 && minor_version >= 3))
+    priv->has_randr13 = TRUE;
+
+#ifdef HAVE_XRANDR15
+  if (major_version > 1 || (major_version == 1 && minor_version >= 5))
+    priv->has_randr15 = TRUE;
+#endif
+
+  init_monitors (manager);
 }
 
 FlashbackMonitorManager *
@@ -1238,11 +1562,11 @@ output_set_underscanning_xrandr (FlashbackMonitorManagerPrivate *priv,
   value = underscanning ? "on" : "off";
   valueatom = XInternAtom (priv->xdisplay, value, False);
 
-  XRRChangeOutputProperty (priv->xdisplay,
-                           (XID)output->winsys_id,
-                           prop,
-                           XA_ATOM, 32, PropModeReplace,
-                           (unsigned char*) &valueatom, 1);
+  xcb_randr_change_output_property (XGetXCBConnection (priv->xdisplay),
+                                    (XID) output->winsys_id,
+                                    prop, XCB_ATOM_ATOM, 32,
+                                    XCB_PROP_MODE_REPLACE,
+                                    1, &valueatom);
 
   /* Configure the border at the same time. Currently, we use a
    * 5% of the width/height of the mode. In the future, we should
@@ -1253,19 +1577,21 @@ output_set_underscanning_xrandr (FlashbackMonitorManagerPrivate *priv,
 
       prop = XInternAtom (priv->xdisplay, "underscan hborder", False);
       border_value = output->crtc->current_mode->width * 0.05;
-      XRRChangeOutputProperty (priv->xdisplay,
-                               (XID)output->winsys_id,
-                               prop,
-                               XA_INTEGER, 32, PropModeReplace,
-                               (unsigned char *) &border_value, 1);
+
+      xcb_randr_change_output_property (XGetXCBConnection (priv->xdisplay),
+                                        (XID) output->winsys_id,
+                                        prop, XCB_ATOM_INTEGER, 32,
+                                        XCB_PROP_MODE_REPLACE,
+                                        1, &border_value);
 
       prop = XInternAtom (priv->xdisplay, "underscan vborder", False);
       border_value = output->crtc->current_mode->height * 0.05;
-      XRRChangeOutputProperty (priv->xdisplay,
-                               (XID)output->winsys_id,
-                               prop,
-                               XA_INTEGER, 32, PropModeReplace,
-                               (unsigned char *) &border_value, 1);
+
+      xcb_randr_change_output_property (XGetXCBConnection (priv->xdisplay),
+                                        (XID) output->winsys_id,
+                                        prop, XCB_ATOM_INTEGER, 32,
+                                        XCB_PROP_MODE_REPLACE,
+                                        1, &border_value);
     }
 }
 
@@ -1390,17 +1716,16 @@ flashback_monitor_manager_apply_configuration (FlashbackMonitorManager  *manager
       if (crtc_info->mode != NULL)
         {
           MetaMonitorMode *mode;
-          XID *outputs;
-          unsigned int j, n_outputs;
-          int width, height;
+          XID *xids;
+          unsigned int j, n_xids;
           Status ok;
 
           mode = crtc_info->mode;
 
-          n_outputs = crtc_info->outputs->len;
-          outputs = g_new (XID, n_outputs);
+          n_xids = crtc_info->outputs->len;
+          xids = g_new (XID, n_xids);
 
-          for (j = 0; j < n_outputs; j++)
+          for (j = 0; j < n_xids; j++)
             {
               MetaOutput *output;
 
@@ -1409,7 +1734,7 @@ flashback_monitor_manager_apply_configuration (FlashbackMonitorManager  *manager
               output->is_dirty = TRUE;
               output->crtc = crtc;
 
-              outputs[j] = output->winsys_id;
+              xids[j] = output->winsys_id;
             }
 
           ok = XRRSetCrtcConfig (priv->xdisplay,
@@ -1419,7 +1744,7 @@ flashback_monitor_manager_apply_configuration (FlashbackMonitorManager  *manager
                                  crtc_info->x, crtc_info->y,
                                  (XID) mode->mode_id,
                                  meta_monitor_transform_to_xrandr (crtc_info->transform),
-                                 outputs, n_outputs);
+                                 xids, n_xids);
 
           if (ok != Success)
             {
@@ -1449,7 +1774,7 @@ flashback_monitor_manager_apply_configuration (FlashbackMonitorManager  *manager
           crtc->transform = crtc_info->transform;
 
         next:
-          g_free (outputs);
+          g_free (xids);
         }
     }
 
@@ -1469,9 +1794,10 @@ flashback_monitor_manager_apply_configuration (FlashbackMonitorManager  *manager
                                       output_info->output,
                                       output_info->is_presentation);
 
-      output_set_underscanning_xrandr (priv,
-                                       output_info->output,
-                                       output_info->is_underscanning);
+      if (output_get_supports_underscanning_xrandr (priv, output_info->output))
+        output_set_underscanning_xrandr (priv,
+                                         output_info->output,
+                                         output_info->is_underscanning);
 
       output->is_primary = output_info->is_primary;
       output->is_presentation = output_info->is_presentation;
@@ -1529,11 +1855,12 @@ flashback_monitor_manager_change_backlight (FlashbackMonitorManager *manager,
   hw_value = round ((double)value / 100.0 * output->backlight_max + output->backlight_min);
 
   atom = XInternAtom (priv->xdisplay, "Backlight", False);
-  XRRChangeOutputProperty (priv->xdisplay,
-                           (XID) output->winsys_id,
-                           atom,
-                           XA_INTEGER, 32, PropModeReplace,
-                           (unsigned char *) &hw_value, 1);
+
+  xcb_randr_change_output_property (XGetXCBConnection (priv->xdisplay),
+                                    (XID) output->winsys_id,
+                                    atom, XCB_ATOM_INTEGER, 32,
+                                    XCB_PROP_MODE_REPLACE,
+                                    1, &hw_value);
 
   /* We're not selecting for property notifies, so update the value immediately */
   output->backlight = normalize_backlight (output, hw_value);
@@ -1599,6 +1926,7 @@ flashback_monitor_manager_set_power_save_mode (FlashbackMonitorManager *manager,
     case META_POWER_SAVE_OFF:
       state = DPMSModeOff;
       break;
+    case META_POWER_SAVE_UNSUPPORTED:
     default:
       return;
   }
@@ -1656,25 +1984,35 @@ meta_output_parse_edid (MetaOutput *meta_output,
   if (parsed_edid)
     {
       meta_output->vendor = g_strndup (parsed_edid->manufacturer_code, 4);
-      if (parsed_edid->dsc_product_name[0])
-        meta_output->product = g_strndup (parsed_edid->dsc_product_name, 14);
-      else
-        meta_output->product = g_strdup_printf ("0x%04x", (unsigned) parsed_edid->product_code);
-      if (parsed_edid->dsc_serial_number[0])
-        meta_output->serial = g_strndup (parsed_edid->dsc_serial_number, 14);
-      else
-        meta_output->serial = g_strdup_printf ("0x%08x", parsed_edid->serial_number);
+      if (!g_utf8_validate (meta_output->vendor, -1, NULL))
+        g_clear_pointer (&meta_output->vendor, g_free);
+
+      meta_output->product = g_strndup (parsed_edid->dsc_product_name, 14);
+      if (!g_utf8_validate (meta_output->product, -1, NULL) ||
+          meta_output->product[0] == '\0')
+        {
+          g_clear_pointer (&meta_output->product, g_free);
+          meta_output->product = g_strdup_printf ("0x%04x", (unsigned) parsed_edid->product_code);
+        }
+
+      meta_output->serial = g_strndup (parsed_edid->dsc_serial_number, 14);
+      if (!g_utf8_validate (meta_output->serial, -1, NULL) ||
+          meta_output->serial[0] == '\0')
+        {
+          g_clear_pointer (&meta_output->serial, g_free);
+          meta_output->serial = g_strdup_printf ("0x%08x", parsed_edid->serial_number);
+        }
 
       g_free (parsed_edid);
     }
 
  out:
   if (!meta_output->vendor)
-    {
-      meta_output->vendor = g_strdup ("unknown");
-      meta_output->product = g_strdup ("unknown");
-      meta_output->serial = g_strdup ("unknown");
-    }
+    meta_output->vendor = g_strdup ("unknown");
+  if (!meta_output->product)
+    meta_output->product = g_strdup ("unknown");
+  if (!meta_output->serial)
+    meta_output->serial = g_strdup ("unknown");
 }
 
 void
@@ -1765,14 +2103,34 @@ void
 flashback_monitor_manager_rebuild_derived (FlashbackMonitorManager *manager)
 {
   MetaMonitorInfo *old_monitor_infos;
+  unsigned old_n_monitor_infos;
+  unsigned i, j;
 
   old_monitor_infos = manager->monitor_infos;
+  old_n_monitor_infos = manager->n_monitor_infos;
 
   if (manager->in_init)
     return;
 
   make_logical_config (manager);
 
+  for (i = 0; i < old_n_monitor_infos; i++)
+    {
+      gboolean delete_mon = TRUE;
+
+      for (j = 0; j < manager->n_monitor_infos; j++)
+        {
+          if (manager->monitor_infos[j].monitor_winsys_xid == old_monitor_infos[i].monitor_winsys_xid)
+            {
+              delete_mon = FALSE;
+              break;
+            }
+        }
+
+      if (delete_mon)
+        remove_monitor (manager, old_monitor_infos[i].monitor_winsys_xid);
+    }
+
   g_signal_emit_by_name (manager->priv->display_config, "monitors-changed");
 
   g_free (old_monitor_infos);
