@@ -41,6 +41,7 @@ typedef struct _MetaMonitorMode MetaMonitorMode;
 typedef struct _MetaMonitorInfo MetaMonitorInfo;
 typedef struct _MetaCRTCInfo    MetaCRTCInfo;
 typedef struct _MetaOutputInfo  MetaOutputInfo;
+typedef struct _MetaTileInfo    MetaTileInfo;
 
 typedef enum {
   META_MONITOR_TRANSFORM_NORMAL,
@@ -74,6 +75,17 @@ typedef enum {
   META_CONNECTOR_TYPE_DSI         = 16,
 } MetaConnectorType;
 
+struct _MetaTileInfo {
+  guint32 group_id;
+  guint32 flags;
+  guint32 max_h_tiles;
+  guint32 max_v_tiles;
+  guint32 loc_h_tile;
+  guint32 loc_v_tile;
+  guint32 tile_w;
+  guint32 tile_h;
+};
+
 struct _MetaOutput
 {
   /* The CRTC driving this output, NULL if the output is not enabled */
@@ -113,6 +125,7 @@ struct _MetaOutput
   gboolean            is_primary;
   gboolean            is_presentation;
   gboolean            is_underscanning;
+  gboolean            supports_underscanning;
 
   gpointer            driver_private;
   GDestroyNotify      driver_notify;
@@ -121,6 +134,8 @@ struct _MetaOutput
   gboolean            hotplug_mode_update;
   gint                suggested_x;
   gint                suggested_y;
+
+  MetaTileInfo        tile_info;
 };
 
 struct _MetaCRTC
@@ -154,6 +169,8 @@ struct _MetaMonitorMode
   GDestroyNotify driver_notify;
 };
 
+#define META_MAX_OUTPUTS_PER_MONITOR 4
+
 /**
  * MetaMonitorInfo:
  *
@@ -169,6 +186,10 @@ struct _MetaMonitorInfo
   int          number;
   int          xinerama_index;
   GdkRectangle rect;
+  /* for tiled monitors these are calculated, from untiled just copied */
+  float        refresh_rate;
+  int          width_mm;
+  int          height_mm;
   gboolean     is_primary;
   gboolean     is_presentation; /* XXX: not yet used */
   gboolean     in_fullscreen;
@@ -182,6 +203,12 @@ struct _MetaMonitorInfo
      the primary one).
   */
   glong        winsys_id;
+
+  guint32      tile_group_id;
+
+  int          monitor_winsys_xid;
+  int          n_outputs;
+  MetaOutput  *outputs[META_MAX_OUTPUTS_PER_MONITOR];
 };
 
 /*
