@@ -35,9 +35,6 @@
 #define KEY_SWITCH_INPUT_SOURCE "switch-input-source"
 #define KEY_SWITCH_INPUT_SOURCE_BACKWARD "switch-input-source-backward"
 
-#define INPUT_SOURCE_TYPE_XKB "xkb"
-#define INPUT_SOURCE_TYPE_IBUS "ibus"
-
 typedef struct _SourceInfo SourceInfo;
 struct _SourceInfo
 {
@@ -73,11 +70,16 @@ struct _GfInputSourceManager
   GList                 *mru_sources_backup;
 
   GtkWidget             *popup;
+
+  GfInputSource         *current_source;
+
+  gboolean               sources_per_window;
 };
 
 enum
 {
   SIGNAL_SOURCES_CHANGED,
+  SIGNAL_CURRENT_SOURCE_CHANGED,
 
   LAST_SIGNAL
 };
@@ -97,6 +99,33 @@ static GParamSpec *properties[LAST_PROP] = { NULL };
 
 G_DEFINE_TYPE (GfInputSourceManager, gf_input_source_manager, G_TYPE_OBJECT)
 
+static void
+change_per_window_source (GfInputSourceManager *manager)
+{
+  if (!manager->sources_per_window)
+    return;
+
+  /* FIXME: */
+}
+
+static gint
+compare_sources_by_index (gconstpointer a,
+                          gconstpointer b)
+{
+  GfInputSource *source1;
+  GfInputSource *source2;
+  guint index1;
+  guint index2;
+
+  source1 = (GfInputSource *) a;
+  source2 = (GfInputSource *) b;
+
+  index1 = gf_input_source_get_index (source1);
+  index2 = gf_input_source_get_index (source2);
+
+  return index1 - index2;
+}
+
 static gchar *
 get_symbol_from_char_code (gunichar code)
 {
@@ -226,10 +255,56 @@ fade_finished_cb (GfInputSourcePopup *popup,
   manager->popup = NULL;
 }
 
+static gint
+compare_indexes (gconstpointer a,
+                 gconstpointer b)
+{
+  return GPOINTER_TO_UINT (a) - GPOINTER_TO_UINT (b);
+}
+
+static gboolean
+modifiers_accelerator_activated_cb (GfKeybindings *keybindings,
+                                    gpointer       user_data)
+{
+  GfInputSourceManager *manager;
+  guint size;
+  GfInputSource *source;
+  GList *keys;
+  guint next_index;
+
+  manager = GF_INPUT_SOURCE_MANAGER (user_data);
+
+  size = g_hash_table_size (manager->input_sources);
+  if (size == 0)
+    {
+      gf_keyboard_manager_ungrab (manager->keyboard_manager, GDK_CURRENT_TIME);
+      return TRUE;
+    }
+
+  source = manager->current_source;
+  if (source == NULL)
+    source = (GfInputSource *) g_hash_table_lookup (manager->input_sources,
+                                                    GUINT_TO_POINTER (0));
+
+  keys = g_hash_table_get_keys (manager->input_sources);
+  keys = g_list_sort (keys, compare_indexes);
+
+  next_index = gf_input_source_get_index (source ) + 1;
+  if (next_index > GPOINTER_TO_UINT (g_list_nth_data (keys, size - 1)))
+    next_index = 0;
+
+  source = (GfInputSource *) g_hash_table_lookup (manager->input_sources,
+                                                  GUINT_TO_POINTER (next_index));
+
+  gf_input_source_activate (source);
+  g_list_free (keys);
+
+  return TRUE;
+}
+
 static void
 accelerator_activated_cb (GfKeybindings *keybindings,
                           guint          action,
-                          GVariant      *parameters,
                           gpointer       user_data)
 {
   GfInputSourceManager *manager;
@@ -266,7 +341,7 @@ static void
 keybindings_init (GfInputSourceManager *manager)
 {
   manager->wm_keybindings = g_settings_new (DESKTOP_WM_KEYBINDINGS_SCHEMA);
-  manager->keybindings = gf_keybindings_new ();
+  manager->keybindings = gf_keybindings_new (TRUE);
 
   g_signal_connect (manager->wm_keybindings,
                     "changed::" KEY_SWITCH_INPUT_SOURCE,
@@ -280,6 +355,8 @@ keybindings_init (GfInputSourceManager *manager)
 
   g_signal_connect (manager->keybindings, "accelerator-activated",
                     G_CALLBACK (accelerator_activated_cb), manager);
+  g_signal_connect (manager->keybindings, "modifiers-accelerator-activated",
+                    G_CALLBACK (modifiers_accelerator_activated_cb), manager);
 
   switch_input_changed_cb (manager->wm_keybindings, NULL, manager);
   switch_input_backward_changed_cb (manager->wm_keybindings, NULL, manager);
@@ -441,8 +518,15 @@ static void
 current_input_source_changed (GfInputSourceManager *manager,
                               GfInputSource        *new_source)
 {
+  GfInputSource *old_source;
   GList *l;
 
+  old_source = manager->current_source;
+  manager->current_source = new_source;
+
+  g_signal_emit (manager, signals[SIGNAL_CURRENT_SOURCE_CHANGED],
+                 0, old_source);
+
   for (l = manager->mru_sources; l != NULL; l = g_list_next (l))
     {
       GfInputSource *source;
@@ -457,6 +541,8 @@ current_input_source_changed (GfInputSourceManager *manager,
           break;
         }
     }
+
+  change_per_window_source (manager);
 }
 
 static void
@@ -506,9 +592,6 @@ activate_cb (GfInputSource *source,
   else
     engine = "xkb:us::eng";
 
-  g_signal_connect (manager->ibus_manager, "engine-set",
-                    G_CALLBACK (engine_set_cb), manager);
-
   gf_ibus_manager_set_engine (manager->ibus_manager, engine);
   current_input_source_changed (manager, source);
 }
@@ -635,7 +718,7 @@ update_mru_sources_list (GfInputSourceManager *manager)
     }
 
   sources = g_hash_table_get_values (manager->input_sources);
-  sources = g_list_reverse (sources);
+  sources = g_list_sort (sources, compare_sources_by_index);
 
   mru_sources = NULL;
   for (l1 = manager->mru_sources; l1 != NULL; l1 = g_list_next (l1))
@@ -698,6 +781,8 @@ sources_changed_cb (GfInputSourceSettings *settings,
   manager->ibus_sources = g_hash_table_new_full (g_str_hash, g_str_equal,
                                                  g_free, g_object_unref);
 
+  manager->current_source = NULL;
+
   sources_by_name = g_hash_table_new_full (g_str_hash, g_str_equal,
                                            g_free, NULL);
 
@@ -778,6 +863,15 @@ static void
 per_window_changed_cb (GfInputSourceSettings *settings,
                        gpointer               user_data)
 {
+  GfInputSourceManager *manager;
+  gboolean per_window;
+
+  manager = GF_INPUT_SOURCE_MANAGER (user_data);
+  per_window = gf_input_source_settings_get_per_window (settings);
+
+  manager->sources_per_window = per_window;
+
+  /* FIXME: */
 }
 
 static void
@@ -791,6 +885,8 @@ input_source_settings_init (GfInputSourceManager *manager)
                     G_CALLBACK (xkb_options_changed_cb), manager);
   g_signal_connect (manager->settings, "per-window-changed",
                     G_CALLBACK (per_window_changed_cb), manager);
+
+  per_window_changed_cb (manager->settings, manager);
 }
 
 static void
@@ -822,6 +918,63 @@ properties_registered_cb (GfIBusManager *ibus_manager,
                           IBusPropList  *prop_list,
                           gpointer       user_data)
 {
+  GfInputSourceManager *manager;
+  GfInputSource *source;
+
+  manager = GF_INPUT_SOURCE_MANAGER (user_data);
+  source = (GfInputSource *) g_hash_table_lookup (manager->ibus_sources,
+                                                  engine_name);
+
+  if (!source)
+    return;
+
+  gf_input_source_set_properties (source, prop_list);
+
+  if (compare_sources (source, manager->current_source))
+    g_signal_emit (manager, signals[SIGNAL_CURRENT_SOURCE_CHANGED], 0, NULL);
+}
+
+static gboolean
+update_sub_property (IBusPropList *prop_list,
+                     IBusProperty *prop)
+{
+  IBusProperty *p;
+  guint index;
+
+  if (!prop_list)
+    return FALSE;
+
+  index = 0;
+  while ((p = ibus_prop_list_get (prop_list, index)) != NULL)
+    {
+      const gchar *p_key;
+      IBusPropType p_type;
+      const gchar *prop_key;
+      IBusPropType prop_type;
+
+      p_key = ibus_property_get_key (p);
+      p_type = ibus_property_get_prop_type (p);
+      prop_key = ibus_property_get_key (prop);
+      prop_type = ibus_property_get_prop_type (prop);
+
+      if (g_strcmp0 (p_key, prop_key) == 0 && p_type == prop_type)
+        {
+          ibus_prop_list_update_property (prop_list, prop);
+          return TRUE;
+        }
+      else if (p_type == PROP_TYPE_MENU)
+        {
+          IBusPropList *sub_props;
+
+          sub_props = ibus_property_get_sub_props (p);
+          if (update_sub_property (sub_props, prop))
+            return TRUE;
+        }
+
+      index++;
+    }
+
+  return FALSE;
 }
 
 static void
@@ -830,6 +983,24 @@ property_updated_cb (GfIBusManager *ibus_manager,
                      IBusProperty  *property,
                      gpointer       user_data)
 {
+  GfInputSourceManager *manager;
+  GfInputSource *source;
+  IBusPropList *prop_list;
+
+  manager = GF_INPUT_SOURCE_MANAGER (user_data);
+  source = (GfInputSource *) g_hash_table_lookup (manager->ibus_sources,
+                                                  engine_name);
+
+  if (!source)
+    return;
+
+  prop_list = gf_input_source_get_properties (source);
+
+  if (!update_sub_property (prop_list, property))
+    return;
+
+  if (compare_sources (source, manager->current_source))
+    g_signal_emit (manager, signals[SIGNAL_CURRENT_SOURCE_CHANGED], 0, NULL);
 }
 
 static void
@@ -895,6 +1066,9 @@ gf_input_source_manager_constructed (GObject *object)
                     G_CALLBACK (property_updated_cb), manager);
   g_signal_connect (manager->ibus_manager, "set-content-type",
                     G_CALLBACK (set_content_type_cb), manager);
+
+  g_signal_connect (manager->ibus_manager, "engine-set",
+                    G_CALLBACK (engine_set_cb), manager);
 }
 
 static void
@@ -1004,6 +1178,11 @@ gf_input_source_manager_class_init (GfInputSourceManagerClass *manager_class)
     g_signal_new ("sources-changed", G_TYPE_FROM_CLASS (manager_class),
                   G_SIGNAL_RUN_LAST, 0, NULL, NULL, NULL, G_TYPE_NONE, 0);
 
+  signals[SIGNAL_CURRENT_SOURCE_CHANGED] =
+    g_signal_new ("current-source-changed", G_TYPE_FROM_CLASS (manager_class),
+                  G_SIGNAL_RUN_LAST, 0, NULL, NULL, NULL, G_TYPE_NONE, 1,
+                  GF_TYPE_INPUT_SOURCE);
+
   properties[PROP_IBUS_MANAGER] =
     g_param_spec_object ("ibus-manager", "IBus Manager",
                          "An instance of GfIBusManager",
@@ -1043,3 +1222,20 @@ gf_input_source_manager_reload (GfInputSourceManager *manager)
 
   sources_changed_cb (manager->settings, manager);
 }
+
+GfInputSource *
+gf_input_source_manager_get_current_source (GfInputSourceManager *manager)
+{
+  return manager->current_source;
+}
+
+GList *
+gf_input_source_manager_get_input_sources (GfInputSourceManager *manager)
+{
+  GList *sources;
+
+  sources = g_hash_table_get_values (manager->input_sources);
+  sources = g_list_sort (sources, compare_sources_by_index);
+
+  return sources;
+}
