@@ -18,13 +18,13 @@
 #ifndef FLASHBACK_LABEL_WINDOW_H
 #define FLASHBACK_LABEL_WINDOW_H
 
-#include <gtk/gtk.h>
+#include <libcommon/gf-popup-window.h>
 
 G_BEGIN_DECLS
 
 #define FLASHBACK_TYPE_LABEL_WINDOW flashback_label_window_get_type ()
 G_DECLARE_FINAL_TYPE (FlashbackLabelWindow, flashback_label_window,
-                      FLASHBACK, LABEL_WINDOW, GtkWindow)
+                      FLASHBACK, LABEL_WINDOW, GfPopupWindow)
 
 FlashbackLabelWindow *flashback_label_window_new  (gint                  monitor,
                                                    const gchar          *label);
